import setuptools

long_description = """
# Coldtype

### Programmatic display typography

More info available at: [coldtype.goodhertz.com](https://coldtype.goodhertz.com)
"""

setuptools.setup(
    name="coldtype",
    version="0.5.2",
    author="Rob Stenson / Goodhertz",
    author_email="rob@goodhertz.com",
    description="Functions for manual vectorized typesetting",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/goodhertz/coldtype",
    #package_dir={"": "coldtype"},
    packages=[
        "coldtype",
        "coldtype.sh",
        "coldtype.fx",
        "coldtype.img",
        "coldtype.time",
        "coldtype.midi",
        "coldtype.pens",
        "coldtype.text",
        "coldtype.grid",
        "coldtype.color",
        "coldtype.capture",
        "coldtype.geometry",
        "coldtype.time.nle",
        "coldtype.renderer",
        "coldtype.renderable",
        "coldtype.webserver",
        "coldtype.fontgoggles",
        "coldtype.interpolation",
        "coldtype.fontgoggles.font",
        "coldtype.fontgoggles.misc",
        "coldtype.fontgoggles.compile",
    ],
    include_package_data=True,
    package_data={
        "": [
            "webserver/webviewer.html",
            "demo/ColdtypeObviously-VF.ttf",
            "demo/demo.py",
        ],
    },
    entry_points={
        'console_scripts': [
            'coldtype = coldtype.renderer:main'
        ],
    },
    extras_require={
        "skia": [
            "skia-python>=86.0",
        ],
        "viewer": [
            "glfw",
            "PyOpenGL",
            "PyOpenGL-accelerate",
            "skia-python>=86.0",
            "skia-pathops", # can this be taken from skia-python?
            "SimpleWebSocketServer",
            "watchdog<2.0.0", # https://github.com/gorakhargosh/watchdog/issues/702
            "noise",
        ],
        "webviewer": [
            "SimpleWebSocketServer",
            "watchdog<2.0.0", # https://github.com/gorakhargosh/watchdog/issues/702
        ],
        "experimental": [
            "pynput",
            "pyaudio",
            "rtmidi",
            "noise",
        ],
        "c": [
            "srt",
            "noise",
        ]
    },
    install_requires=[
        # drafting
        "lxml",
        "fonttools[ufo]",
        "fontPens",
        "more-itertools",
        "easing-functions",
        "timecode",
        # /drafting
        "mido",
        "numpy",
        "defcon",
        "docutils",
        # drafting-text
        "freetype-py",
        "uharfbuzz>=0.14.0",
        "unicodedata2",
        "ufo2ft",
        "python-bidi",
        # /drafting-text
        "more-itertools",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



import re

from coldtype.pens.draftingpen import DraftingPen
from coldtype.pens.draftingpens import DraftingPens
from coldtype.text.composer import Graf, GrafStyle, Lockup
from coldtype.text.reader import StyledString, Style
from coldtype.color import hsl

from pathlib import PurePath
from typing import Optional, List, Callable, Tuple, Union
from functools import reduce

try:
    from pygments import highlight
    from pygments.lexers import PythonLexer
    from pygments.formatter import Formatter
    from pygments.token import Token
except ImportError:
    highlight = None
    pass

class RichText(DraftingPens):
    """Very experimental module to support rich-text from annotated strings, like a super-minimal-but-open-ended subset of markdown, inspired by the way rich text is built up in the time.nle.premiere DPS subclass in coldtype"""
    def __init__(self,
        rect,
        text,
        render_text_fn:Union[dict, Callable[[str, List[str]], Tuple[str, Style]]],
        fit=None,
        graf_style=GrafStyle(leading=20),
        tag_delimiters=["[", "]"],
        visible_boundaries=[" "],
        invisible_boundaries=[],
        union_styles=True,
        blankfill="¶",
        strip=True,
        strip_lines=False):
        super().__init__()
        self.tag_delimiters = tag_delimiters
        self.visible_boundary_chars = visible_boundaries
        self.invisible_boundary_chars = invisible_boundaries
        self.union_styles = union_styles
        self.blankfill = blankfill
        
        if isinstance(text, PurePath):
            text = text.read_text()
        if strip:
            text = text.strip()
        
        if strip_lines:
            text = "\n".join([l.strip() for l in text.split("\n")])
        
        self._pens = self.parse_block(text, render_text_fn, rect, fit, graf_style)._pens

    def parse_block(self, txt, render_text_fn, rect, fit, graf_style):
        parsed_lines = []
        alt_parsed_lines = []

        for line in txt.splitlines():
            line_meta = []
            line_result = []

            i = 0
            in_tag = False
            bnd_char = ""
            si = 0
            slugs = {0:""}
            metas = {0:""}
            bnds = {0:""}

            while i < len(line):
                if in_tag:
                    if line[i] == self.tag_delimiters[1]:
                        in_tag = False
                    else:
                        metas[si] += line[i]
                elif line[i] == self.tag_delimiters[0]:
                    in_tag = True
                
                elif line[i] in self.visible_boundary_chars or line[i] in self.invisible_boundary_chars:
                    bnd_char = line[i]
                    bnds[si] = bnd_char
                    si += 1
                    slugs[si] = ""
                    metas[si] = ""
                    bnds[si] = ""
                else:
                    slugs[si] += line[i]
                i += 1
            
            if not list(slugs.values())[-1]:
                line_meta = list(metas.values())[-1]
                last_key = list(slugs.keys())[-1]
                del slugs[last_key]
                bnds[si-1] = ""
            else:
                line_meta = []
            
            alt_line_result = []
            for i, slug in slugs.items():
                b = bnds[i]
                if b in self.invisible_boundary_chars:
                    b = ""
                styles = []
                if metas[i]:
                    styles.append(metas[i])
                if line_meta:
                    styles.append(line_meta)
                alt_line_result.append([slug + b, styles])
            alt_parsed_lines.append(alt_line_result)

        parsed_lines = []
        for pl in alt_parsed_lines:
            #print(">", pl)
            if pl:
                parsed_lines.append(pl)
            elif self.blankfill:
                parsed_lines.append([[self.blankfill, ["blank"]]])

        lines = []
        groupings = []

        #from pprint import pprint
        #pprint(parsed_lines)

        for idx, line in enumerate(parsed_lines):
            slugs = []
            texts = []
            for txt, styles in line:
                if callable(render_text_fn):
                    ftxt, style = render_text_fn(txt, styles)
                else:
                    style = styles[0] if len(styles) > 0 else "default"
                    ftxt = txt
                    if style in render_text_fn:
                        style = render_text_fn[style]
                    else:
                        style = render_text_fn["default"]
                texts.append([ftxt, idx, style, styles])

            grouped_texts = []
            idx = 0
            done = False
            while not done:
                style = texts[idx][2]
                grouped_text = [texts[idx]]
                style_same = True
                while style_same:
                    idx += 1
                    try:
                        next_style = texts[idx][2]
                        if next_style == style and self.union_styles:
                            style_same = True
                            grouped_text.append(texts[idx])
                        else:
                            style_same = False
                            grouped_texts.append(grouped_text)
                    except IndexError:
                        done = True
                        style_same = False
                        grouped_texts.append(grouped_text)

            for gt in grouped_texts:
                full_text = "".join([t[0] for t in gt])
                styles = gt[0][3]
                style = gt[0][2].mod()
                style.data = dict(style_names=styles, txt=full_text)
                slugs.append(StyledString(full_text, style))
            groupings.append(grouped_texts)
            lines.append(slugs)

        lockups = []
        for line in lines:
            lockup = Lockup(line, preserveLetters=True, nestSlugs=True)
            if fit:
                lockup.fit(fit)
            lockups.append(lockup)
        graf = Graf(lockups, rect, graf_style)
        pens = graf.pens()#.align(rect, x="minx")
        group_pens = DraftingPens() # TODO configurable?

        pens.reversePens()
        for line in pens:
            line.reversePens()
            for slug in line:
                slug.reversePens()
        return pens
    
    def filter_style(self, style):
        return self.pfilter(lambda i, p: style in p.data.get("style_names", []))
    
    def filter_text(self, text, flags=re.I):
        return self.pfilter(lambda i, p: re.match(text, p.data.get("txt", ""), flags=flags))
    
    def remove_blanklines(self, blank=None):
        if not blank and self.blankfill:
            blank = self.blankfill
        
        for line in self._pens:
            txt = reduce(lambda acc, p: p.data.get("txt", "") + acc, line, "")
            if txt == blank:
                line._pens = [DraftingPen()]
        
        return self


if highlight:

    class ColdtypeFormatter(Formatter):
        def format(self, tokensource, outfile):
            for ttype, token in tokensource:
                if ttype != Token.Text:
                    tt = re.sub(r"^Token\.", "", str(ttype))
                    outfile.write(f"¬{token}≤{tt}≥")
                else:
                    outfile.write(token)
    

    class PythonCode(RichText):
        def __init__(self,
            rect,
            text,
            render_text_fn:Callable[[str, List[str]], Tuple[str, Style]],
            fit=None,
            graf_style=GrafStyle(leading=20)):

            if isinstance(text, PurePath):
                text = text.read_text()

            txt = highlight(text, PythonLexer(), ColdtypeFormatter())
            super().__init__(
                rect, txt, render_text_fn,
                fit=fit, graf_style=graf_style,
                tag_delimiters=["≤", "≥"],
                visible_boundaries=[],
                invisible_boundaries=["¬"])
        
        def DefaultStyles(r, b, bi):
            """regular, bold, bold-italic"""
            return {
                "Keyword": (bi, hsl(0.9, s=0.6)),
                "Keyword.Namespace": (b, hsl(0.2, s=1)),
                "Name.Namespace": (b, hsl(0.55, s=1)),
                "Name.Builtin": (bi, hsl(0.05, s=1)),
                "Name.Function": (bi, hsl(0.65, s=1, l=0.7)),
                "Name.Decorator": (bi, hsl(0.2, 1)),
                "Name": (b, hsl(0.45, 0.7)),
                "Operator": (b, hsl(0.6, s=1)),
                "Operator.Word": (b, hsl(0.9)),
                "Punctuation": (b, hsl(0.8)),
                "Literal.String.Double": (r, hsl(0.15, s=0.7)),
                "Literal.String.Affix": (bi, hsl(0.85, s=1)),
                "Literal.String.Interpol": (b, hsl(0.05, s=1)),
                "Literal.String.Doc": (r, hsl(0.6, l=0.4)),
                "Literal.Number.Float": (r, hsl(0.9, s=0.7)),
                "Literal.Number.Integer": (r, hsl(0.45)),
                "Comment.Single": (r, (0.3)),
                "default": (r, 0.5),
            }



from coldtype.time.timeable import Timeable


class Timeline(Timeable):
    """
    General base class for any kind of timeline or sequence, in the NLE sense

    * ``fps`` is `frames-per-second`, and defaults to 30
    * ``storyboard`` is useful if you want to always see a given frame when your animation loads for the first time, or you'd like to see multiple frames, i.e. ``storyboard=[0, 5]`` to see both frame 0 and frame 5 in the viewer
    * ``tracks`` is mostly used internally by subclasses that process external data from another NLE or data source like subtitles
    * ``jumps`` can be used to mark "jump-points" within a timeline, for easy skipping around with the up/down arrows (akin to up/down arrow movements in NLEs)
    """

    __name__ = "Generic"

    def __init__(self, duration, fps=30, storyboard=None, tracks=None, jumps=None):
        self.fps = fps
        self.start = 0
        self.end = duration
        self.tracks = tracks or []
        self._jumps = [self.start, *(jumps or []), self.end-1]
        
        if not storyboard:
            self.storyboard = [0]
        else:
            self.storyboard = storyboard
        if len(self.storyboard) == 0:
            self.storyboard.append(0)
        self.storyboard.sort()
    
    def jumps(self):
        return self._jumps
    
    def text_for_frame(self, fi):
        return ""
    
    def __str__(self):
        return "<coldtype.time.timeline({:s}):{:04d}f@{:02.2f}fps[{:s}]>".format(self.__name__, self.duration, self.fps, ",".join([str(i) for i in self.storyboard]))
    
    def __getitem__(self, item):
        if isinstance(item, str):
            for t in self.tracks:
                if hasattr(t, "name") and t.name == item:
                    return t
        else:
            return self.tracks[item]


from enum import Enum

try:
    import glfw
except ImportError:
    glfw = None

class KeyboardShortcut(Enum):
    PreviewPrev = "prev_prev"
    PreviewPrevMany = "prev_prev_many"
    PreviewNext = "prev_next"
    PreviewNextMany = "prev_next_many"
    
    ClearLastRender = "clear_last_render"
    ClearRenderedFrames = "clear_rendered_frames"
    
    PlayRendered = "play_rendered"
    PlayPreview = "play_preview"
    PlayPreviewSlow = "play_preview_slow"
    PlayAudioFrame = "play_audio_frame"

    ReloadSource = "reload_source"
    RestartApp = "restart_app"
    Quit = "quit"
    Kill = "kill"

    Build = "build"
    Release = "release"
    RenderAll = "render_all"
    RenderWorkarea = "render_workarea"
    ToggleMultiplex = "toggle_multiplex"

    SetWorkareaIn = "set_workarea_in"
    SetWorkareaOut = "set_workarea_out"
    
    JumpPrev = "jump_prev"
    JumpNext = "jump_next"
    JumpHome = "jump_home"
    JumpEnd = "jump_end"
    JumpStoryboard = "jump_storyboard"

    KeylayerEditing = "keylayer_editing"
    KeylayerCmd = "keylayer_command"
    KeylayerText = "keylayer_text"
    OverlayInfo = "overlay_info"
    OverlayTimeline = "overlay_timeline"

    PreviewScaleDown = "preview_scale_down"
    PreviewScaleUp = "preview_scale_up"
    PreviewScaleMin = "preview_scale_min"
    PreviewScaleMax = "preview_scale_max"
    PreviewScaleDefault = "preview_scale_default"

    WindowOpacityDown = "window_opacity_down"
    WindowOpacityUp = "window_opacity_up"
    WindowOpacityMin = "window_opacity_min"
    WindowOpacityMax = "window_opacity_max"

    MIDIControllersPersist = "midi_controllers_persist"
    MIDIControllersClear = "midi_controllers_clear"
    MIDIControllersReset = "midi_controllers_reset"

    JumpToFrameFunctionDef = "jump_to_function_def"

    ViewerTakeFocus = "viewer_take_focus"

    ViewerSoloNone = "viewer_solo_none"
    ViewerSoloNext = "viewer_solo_next",
    ViewerSoloPrev = "viewer_solo_prev",
    ViewerSolo1 = "viewer_solo_1"
    ViewerSolo2 = "viewer_solo_2"
    ViewerSolo3 = "viewer_solo_3"
    ViewerSolo4 = "viewer_solo_4"
    ViewerSolo5 = "viewer_solo_5"
    ViewerSolo6 = "viewer_solo_6"
    ViewerSolo7 = "viewer_solo_7"
    ViewerSolo8 = "viewer_solo_8"
    ViewerSolo9 = "viewer_solo_9"

    ToggleCapturing = "toggle_capturing"
    CaptureOnce = "capture_once"

REPEATABLE_SHORTCUTS = [
    KeyboardShortcut.PreviewPrev,
    KeyboardShortcut.PreviewPrevMany,
    KeyboardShortcut.PreviewNext,
    KeyboardShortcut.PreviewNextMany,
    KeyboardShortcut.PreviewScaleUp,
    KeyboardShortcut.PreviewScaleDown,
    KeyboardShortcut.JumpPrev,
    KeyboardShortcut.JumpNext
]

def symbol_to_glfw(s):
    lookup = {
        "cmd": glfw.MOD_SUPER,
        "ctrl": glfw.MOD_CONTROL,
        "shift": glfw.MOD_SHIFT,
        "<up>": glfw.KEY_UP,
        "<down>": glfw.KEY_DOWN,
        "<left>": glfw.KEY_LEFT,
        "<right>": glfw.KEY_RIGHT,
        "<space>": glfw.KEY_SPACE,
        "<home>": glfw.KEY_HOME,
        "<end>": glfw.KEY_END,
        "<enter>": glfw.KEY_ENTER,
        "-": glfw.KEY_MINUS,
        "=": glfw.KEY_EQUAL,
        "/": glfw.KEY_SLASH,
        "<backslash>": glfw.KEY_BACKSLASH,
    }
    if s in lookup:
        return lookup[s]
    else:
        k = f"KEY_{s.upper()}"
        if hasattr(glfw, k):
            return getattr(glfw, k)
        elif "np" in s:
            return getattr(glfw, f"KEY_KP_{s[2:]}")
        else:
            raise Exception("Invalid keyboard shortcut symbol", s)

SHORTCUTS = {
    KeyboardShortcut.PreviewPrevMany: [
        [["shift"], "j"],
        [["shift"], "<left>"]
    ],
    KeyboardShortcut.PreviewPrev: [
        [[], "j"],
        [[], "<left>"]
    ],
    KeyboardShortcut.PreviewNextMany: [
        [["shift"], "l"],
        [["shift"], "<right>"]
    ],
    KeyboardShortcut.PreviewNext: [
        [[], "l"],
        [[], "<right>"]
    ],
    
    KeyboardShortcut.ClearLastRender: [
        [[], "<backslash>"]
    ],
    KeyboardShortcut.ClearRenderedFrames: [
        [["shift"], "<backslash>"]
    ],
    
    KeyboardShortcut.PlayRendered: [
        [["shift"], "<space>"],
        #[["shift"], glfw.KEY_K]
    ],
    KeyboardShortcut.PlayPreview: [
        [[], "<space>"],
        #[[], glfw.KEY_K]
    ],
    KeyboardShortcut.PlayAudioFrame: [
        [[], "h"]
    ],

    KeyboardShortcut.ReloadSource: [
        [[], "<enter>"],
        [[], "p"],
    ],
    KeyboardShortcut.RestartApp: [
        [["cmd"], "r"]
    ],
    KeyboardShortcut.Quit: [
        [[], "q"]
    ],

    KeyboardShortcut.Release: [
        [[], "r"],
    ],
    KeyboardShortcut.Build: [
        [[], "b"],
    ],
    KeyboardShortcut.RenderAll: [
        [[], "a"],
    ],
    KeyboardShortcut.RenderWorkarea: [
        [[], "w"]
    ],
    KeyboardShortcut.ToggleMultiplex: [
        [[], "m"]
    ],

    KeyboardShortcut.SetWorkareaIn: [
        [["cmd"], "i"]
    ],
    KeyboardShortcut.SetWorkareaOut: [
        [["cmd"], "o"]
    ],

    KeyboardShortcut.JumpPrev: [
        [[], "<up>"],
        [[], "i"]
    ],
    KeyboardShortcut.JumpNext: [
        [[], "<down>"],
        [[], "k"]
    ],
    KeyboardShortcut.JumpHome: [
        [[], "<home>"],
        [["shift"], "h"]
    ],
    KeyboardShortcut.JumpEnd: [
        [[], "<end>"]
    ],
    KeyboardShortcut.JumpStoryboard: [
        [["cmd"], "<home>"]
    ],

    KeyboardShortcut.KeylayerEditing: [
        [[], "d"]
    ],
    KeyboardShortcut.KeylayerCmd: [
        [[], "c"],
    ],
    KeyboardShortcut.KeylayerText: [
        [[], "t"],
    ],

    KeyboardShortcut.OverlayInfo: [
        [[], "/"]
    ],
    KeyboardShortcut.OverlayTimeline: [
        [["cmd"], "t"]
    ],

    KeyboardShortcut.PreviewScaleUp: [
        [[], "="]
    ],
    KeyboardShortcut.PreviewScaleDown: [
        [[], "-"]
    ],
    KeyboardShortcut.PreviewScaleMin: [
        [["cmd"], "-"]
    ],
    KeyboardShortcut.PreviewScaleMax: [
        [["cmd"], "="]
    ],
    KeyboardShortcut.PreviewScaleDefault: [
        [["cmd"], "0"]
    ],

    KeyboardShortcut.WindowOpacityDown: [
        [["cmd"], "<down>"]
    ],
    KeyboardShortcut.WindowOpacityUp: [
        [["cmd"], "<up>"]
    ],
    KeyboardShortcut.WindowOpacityMin: [
        [["cmd", "shift"], "<down>"],
    ],
    KeyboardShortcut.WindowOpacityMax: [
        [["cmd", "shift"], "<up>"]
    ],

    KeyboardShortcut.JumpToFrameFunctionDef: [
        [[], "f"],
    ],

    KeyboardShortcut.ViewerSoloNone: [
        [[], "np0"],
        [[], "0"]
    ],
    KeyboardShortcut.ViewerSoloNext: [
        [["cmd"], "<right>"]
    ],
    KeyboardShortcut.ViewerSoloPrev: [
        [["cmd"], "<left>"]
    ],
    KeyboardShortcut.ViewerSolo1: [
        [[], "np1"],
        [[], "1"]
    ],
    KeyboardShortcut.ViewerSolo2: [
        [[], "np2"],
        [[], "2"]
    ],
    KeyboardShortcut.ViewerSolo3: [
        [[], "np3"],
        [[], "3"]
    ],
    KeyboardShortcut.ViewerSolo4: [
        [[], "np4"],
        [[], "4"]
    ],
    KeyboardShortcut.ViewerSolo5: [
        [[], "np5"],
        [[], "5"]
    ],
    KeyboardShortcut.ViewerSolo6: [
        [[], "np6"],
        [[], "6"]
    ],
    KeyboardShortcut.ViewerSolo7: [
        [[], "np7"],
        [[], "7"]
    ],
    KeyboardShortcut.ViewerSolo8: [
        [[], "np8"],
        [[], "8"]
    ],
    KeyboardShortcut.ViewerSolo9: [
        [[], "np9"],
        [[], "9"]
    ],

    KeyboardShortcut.ToggleCapturing: [
        [["shift"], "c"]
    ],
    KeyboardShortcut.CaptureOnce: [
        [["cmd", "shift"], "c"]
    ]
}


import math
from coldtype.beziers import CurveCutter, splitCubicAtT


def bend(pen, curve, tangent=True):
    def _bend(pen):
        cc = CurveCutter(curve)
        ccl = cc.length
        dpl = pen.bounds().point("SE").x
        xf = ccl/dpl

        def bender(x, y):
            p, tan = cc.subsegmentPoint(end=x*xf)
            px, py = p
            if tangent:
                a = math.sin(math.radians(180+tan)) * y
                b = math.cos(math.radians(180+tan)) * y
                return (px+a, py+b)
                #return (px, y+py)
            else:
                return (px, y+py)
        return pen.nonlinear_transform(bender)
    return _bend


def bend2(curve, tangent=True, offset=(0, 1)):
    def _bend(pen):
        bw = pen.bounds().w
        a = curve.value[0][-1][0]
        b, c, d = curve.value[1][-1]
        def bender(x, y):
            c1, c2 = splitCubicAtT(a, b, c, d, offset[0] + (x/bw)*offset[1])
            _, _a, _b, _c = c1
            if tangent:
                tan = math.degrees(math.atan2(_c[1] - _b[1], _c[0] - _b[0]) + math.pi*.5)
                ax = math.sin(math.radians(90-tan)) * y
                by = math.cos(math.radians(90-tan)) * y
                return _c[0]+ax, (y+_c[1])+by
            return _c[0], y+_c[1]
        return pen.nonlinear_transform(bender)
    return _bend


def bend3(curve, tangent=False, offset=(0, 1)):
    def _bend(pen):
        a = curve.value[0][-1][0]
        b, c, d = curve.value[1][-1]
        bh = pen.bounds().h
        
        def bender(x, y):
            c1, c2 = splitCubicAtT(a, b, c, d, offset[0] + (y/bh)*offset[1])
            _, _a, _b, _c = c1
            if tangent:
                tan = math.degrees(math.atan2(_c[1] - _b[1], _c[0] - _b[0]) + math.pi*.5)
                ax = math.sin(math.radians(90-tan)) * y
                by = math.cos(math.radians(90-tan)) * y
                return x+_c[0]+ax, (y+_c[1])+by
            return x+_c[0], _c[1]
        return pen.nonlinear_transform(bender)
    return _bend


import json
from coldtype.geometry import Rect

def _parse(el, depth=0, chain=False):
    out = []
    chaining = False

    for idx, e in enumerate(el):
        if idx == 0:
            if e == "ø":
                return ""
            elif e == "R":
                out.append("Rect")
            elif e == "P":
                out.append("DATPen")
            elif e == "Ps":
                out.append("DATPens")
            elif e == "S":
                out.append("StSt")
            elif e in ["hsl", "rgb"]: # general whitelist?
                out.append(e)
            else:
                if not chain:
                    out.append(f".{e}")
                else:
                    t = "    " * depth
                    out.append(f"\n{t}.{e}")
            out.append("(")
        elif e == ".":
            chaining = True
            out.append(")")
        elif isinstance(e, str) or isinstance(e, int) or isinstance(e, float):
            if e == "®":
                e = "r"
            elif isinstance(e, str) and "=" not in e and not e.startswith("."):
                if "\n" in e:
                    e = f"\"\"\"{e}\"\"\""
                else:
                    e = f"\"{e}\""
            
            if isinstance(e, str) and e.startswith("."):
                out[-1] += e
            elif idx < 2:
                out.append(str(e))
            else:
                if out[0] == "StSt" and idx == 2:
                    out.append(f", font_cache.get({e}, DEFAULT_FONT)")
                else:
                    out.append(f", {e}")
        elif isinstance(e, dict):
            out.append(f", **{e}")
        elif e is None:
            out.append(None)
        else:
            parsed = _parse(e, depth+1, out[0] == "DATPen" or out[0] == "DATPens" or out[0] == "StSt")
            parsed = "".join([str(el) for el in parsed])
            out.append(parsed)

    if not chaining:
        out.append(")")
    return "".join([str(el) for el in out])


def parse(tree):
    if isinstance(tree, str):
        data = json.loads(tree)
    else:
        data = json.loads(json.dumps(tree))
    
    out = []
    for el in data:
        out.append("(" + _parse(el) + ")")

    to_eval = "[" + ",\n ".join(out) + "]"
    return to_eval

def evalcdel(tree,
    r=Rect(1080, 1080),
    font_cache={},
    DEFAULT_FONT=None):
    from coldtype.color import hsl, rgb
    from coldtype.pens.datpen import DATPen, DATPens
    from coldtype.text import StSt, StyledString, Style
    return eval(parse(tree))

if __name__ == "__main__":
    test = [
        ["P", ".",
            ["oval", ["R", 500, 500, ".", ["inset", 50]]],
            ["f", ["hsl", 0.3, "a=0.5"]],
            ["align", "®"]
        ],
        ["S", "COOL", "ColdtypeObviously", 12, {"wdth":0.5},
            ".",
            ["pen"],
            ["align", "®"]]
    ]

    print(evalcdel(test))


import inspect, platform, re, tempfile, math, datetime

from coldtype.pens.draftingpens import DraftingPens

try:
    import skia
    from coldtype.pens.skiapen import SkiaPen
except ImportError:
    skia = None
    SkiaPen = None

from enum import Enum
from subprocess import run
from pathlib import Path

from coldtype.geometry import Rect, Point
from coldtype.color import normalize_color
from coldtype.text.reader import normalize_font_prefix, Font
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.pens.dattext import DATText
from coldtype.img.datimage import DATImage

try:
    import drawBot as db
    import AppKit
except ImportError:
    db = None

class Keylayer(Enum):
    Default = 0
    Cmd = 1
    Editing = 2
    Text = 3

class Overlay(Enum):
    Info = "info"
    Timeline = "timeline"

class Action(Enum):
    Initial = "initial"
    Resave = "resave"
    RenderAll = "render_all"
    RenderWorkarea = "render_workarea"
    RenderIndices = "render_indices"
    Build = "build"
    Release = "release"
    PreviewStoryboard = "preview_storyboard"
    PreviewStoryboardReload = "preview_storyboard_reload"
    PreviewPlay = "preview_play"
    PreviewIndices = "preview_indices"
    PreviewStoryboardNext = "preview_storyboard_next"
    PreviewStoryboardPrev = "preview_storyboard_prev"
    PreviewStoryboardNextMany = "preview_storyboard_next_many"
    PreviewStoryboardPrevMany = "preview_storyboard_prev_many"
    ClearLastRender = "clear_last_render"
    ClearRenderedFrames = "clear_rendered_frames"
    RenderedPlay = "rendered_play"
    ArbitraryTyping = "arbitrary_typing"
    ArbitraryCommand = "arbitrary_command"
    RestartRenderer = "restart_renderer"
    ToggleMultiplex = "toggle_multiplex"
    Kill = "kill"


class RenderPass():
    def __init__(self, render, suffix, args):
        self.render = render
        self.fn = self.render.func
        self.args = args
        self.suffix = suffix
        self.path = None
        self.output_path = None
    
    def __repr__(self):
        return f"<RenderPass:f{self.output_path}/>"


class renderable():
    """
    Base class for any content renderable by Coldtype
    """
    def __init__(self,
        rect=(1080, 1080),
        bg="whitesmoke",
        fmt="png",
        name=None,
        rasterizer=None,
        prefix=None,
        suffix=None,
        dst=None,
        custom_folder=None,
        postfn=None,
        watch=[],
        watch_restarts=[],
        watch_soft=[],
        solo=False,
        rstate=False,
        preview_only=False,
        direct_draw=False,
        clip=False,
        composites=False,
        cv2caps=None,
        bg_render=False,
        style="default",
        viewBox=True,
        layer=False):
        """Base configuration for a renderable function"""

        self.rect = Rect(rect).round()
        self.bg = normalize_color(bg)
        self.fmt = fmt
        self.prefix = prefix
        self.suffix = suffix
        self.dst = Path(dst).expanduser().resolve() if dst else None
        self.custom_folder = custom_folder
        self.postfn = postfn
        self.last_passes = []
        self.last_result = None
        self.style = style
        self.composites = composites
        self.cv2caps = cv2caps

        self.watch = []
        for w in watch:
            self.add_watchee(w)

        self.watch_restarts = []
        for w in watch_restarts:
            self.watch_restarts.append(self.add_watchee(w, "restart"))
        
        self.watch_soft = []
        for w in watch_soft:
            self.watch_soft.append(self.add_watchee(w, "soft"))

        self.name = name
        self.codepath = None
        self.rasterizer = rasterizer
        self.self_rasterizing = False
        self.hidden = solo == -1
        self.solo = solo
        self.preview_only = preview_only
        self.rstate = rstate
        self.clip = clip
        self.viewBox = viewBox
        self.direct_draw = direct_draw
        self.bg_render = bg_render
        self.layer = layer
        if self.layer:
            self.bg = normalize_color(None)

        if not rasterizer:
            if self.fmt == "svg":
                self.rasterizer = "svg"
            elif self.fmt == "pickle":
                self.rasterizer = "pickle"
            else:
                self.rasterizer = "skia"
    
    def add_watchee(self, w, flag=None):
        try:
            pw = Path(w).expanduser().resolve()
            if not pw.exists():
                print(w, "<<< does not exist (cannot be watched)")
            else:
                self.watch.append([pw, flag])
                return pw
        except TypeError:
            if isinstance(w, Font):
                self.watch.append([w, flag])
            else:
                raise Exception("Can only watch path strings, Paths, and Fonts")
    
    def __call__(self, func):
        self.func = func
        if not self.name:
            self.name = self.func.__name__
        return self
    
    def folder(self, filepath):
        return ""
    
    def pass_suffix(self):
        return self.name
    
    def passes(self, action, renderer_state, indices=[]):
        return [RenderPass(self, self.pass_suffix(), [self.rect])]

    def package(self, filepath, output_folder):
        pass

    def run(self, render_pass, renderer_state):
        if self.rstate:
            return render_pass.fn(*render_pass.args, renderer_state)
        else:
            return render_pass.fn(*render_pass.args)
    
    def runpost(self, result, render_pass, renderer_state):
        if self.postfn:
            return self.postfn(self, result)
        else:
            return result
    
    def draw_preview(self, scale, canvas, rect, result, render_pass): # canvas:skia.Canvas
        sr = self.rect.scale(scale, "mnx", "mxx")
        SkiaPen.CompositeToCanvas(result, sr, canvas, scale, style=self.style)
    
    def hide(self):
        self.hidden = True
        return self
    
    def show(self):
        self.hidden = False
        return self
    
    def normalize_result(self, pens):
        if not pens:
            return DATPens()
        elif hasattr(pens, "_pens"):
            if (isinstance(pens, DraftingPens)
                and not isinstance(pens, DATPens)):
                return DATPens(pens._pens)
            return pens
        elif isinstance(pens, DATPen):
            return DATPens([pens])
        elif isinstance(pens, DATText):
            return DATPens([pens])
        elif isinstance(pens, DATImage):
            return DATPens([pens])
        elif not isinstance(pens, DATPens):
            return DATPens(pens)
        else:
            return pens


class skia_direct(renderable):
    def __init__(self, rect=(1080, 1080), **kwargs):
        super().__init__(rect=rect, direct_draw=True, **kwargs)
    
    def run(self, render_pass, renderer_state, canvas):
        if self.rstate:
            return render_pass.fn(*render_pass.args, renderer_state, canvas)
        else:
            return render_pass.fn(*render_pass.args, canvas)


class drawbot_script(renderable):
    def __init__(self, rect=(1080, 1080), scale=1, **kwargs):
        if not db:
            raise Exception("DrawBot not installed!")
        super().__init__(rect=Rect(rect).scale(scale), rasterizer="drawbot", **kwargs)
        self.self_rasterizing = True
    
    def normalize_result(self, pens):
        return pens
    
    def run(self, render_pass, renderer_state):
        from coldtype.pens.drawbotpen import DrawBotPen
        use_pool = True
        if use_pool:
            pool = AppKit.NSAutoreleasePool.alloc().init()
        try:
            db.newDrawing()
            if renderer_state.previewing:
                ps = renderer_state.preview_scale
                db.size(self.rect.w*ps, self.rect.h*ps)
                db.scale(ps, ps)
                DATPen().rect(self.rect).f(self.bg).cast(DrawBotPen).draw()
            else:
                db.size(self.rect.w, self.rect.h)
            if self.rstate:
                render_pass.fn(*render_pass.args, renderer_state)
            else:
                render_pass.fn(*render_pass.args)
            result = None
            if renderer_state.previewing:
                previews = (render_pass.output_path.parent / "_previews")
                previews.mkdir(exist_ok=True, parents=True)
                preview_frame = previews / render_pass.output_path.name
                db.saveImage(str(preview_frame))
                result = preview_frame
            else:
                render_pass.output_path.parent.mkdir(exist_ok=True, parents=True)
                db.saveImage(str(render_pass.output_path))
                result = render_pass.output_path
            db.endDrawing()
        finally:
            if use_pool:
                del pool
        return result


class svgicon(renderable):
    def __init__(self, **kwargs):
        super().__init__(fmt="svg", **kwargs)
    
    def folder(self, filepath):
        return filepath.stem


class glyph(renderable):
    def __init__(self, glyphName, width=500, **kwargs):
        r = Rect(kwargs.get("rect", Rect(1000, 1000)))
        kwargs.pop("rect", None)
        self.width = width
        self.body = r.take(750, "mdy").take(self.width, "mdx")
        self.glyphName = glyphName
        super().__init__(rect=r, **kwargs)
    
    def passes(self, action, renderer_state, indices=[]):
        return [RenderPass(self, self.glyphName, [])]


class fontpreview(renderable):
    def __init__(self, font_dir, font_re, rect=(1200, 150), limit=25, **kwargs):
        super().__init__(rect=rect, **kwargs)
        self.dir = normalize_font_prefix(font_dir)
        self.re = font_re
        self.matches = []
        
        for font in self.dir.iterdir():
            if re.search(self.re, str(font)):
                if len(self.matches) < limit:
                    self.matches.append(font)
        
        self.matches.sort()
    
    def passes(self, action, renderer_state, indices=[]):
        return [RenderPass(self, "{:s}".format(m.name), [self.rect, m]) for m in self.matches]


class iconset(renderable):
    valid_sizes = [16, 32, 64, 128, 256, 512, 1024]

    def __init__(self, sizes=[128, 1024], **kwargs):
        super().__init__(**kwargs)
        self.sizes = sizes
    
    def folder(self, filepath):
        return f"{filepath.stem}_source"
    
    def passes(self, action, renderer_state, indices=[]): # TODO could use the indices here
        sizes = self.sizes
        if action == Action.RenderAll:
            sizes = self.valid_sizes
        return [RenderPass(self, str(size), [self.rect, size]) for size in sizes]
    
    def package(self, filepath, output_folder):
        # inspired by https://retifrav.github.io/blog/2018/10/09/macos-convert-png-to-icns/
        iconset = output_folder.parent / f"{filepath.stem}.iconset"
        iconset.mkdir(parents=True, exist_ok=True)

        system = platform.system()
        
        if system == "Darwin":
            for png in output_folder.glob("*.png"):
                d = int(png.stem.split("_")[1])
                for x in [1, 2]:
                    if x == 2 and d == 16:
                        continue
                    elif x == 1:
                        fn = f"icon_{d}x{d}.png"
                    elif x == 2:
                        fn = f"icon_{int(d/2)}x{int(d/2)}@2x.png"
                    print(fn)
                run(["sips", "-z", str(d), str(d), str(png), "--out", str(iconset / fn)])
            run(["iconutil", "-c", "icns", str(iconset)])
        
        if True: # can be done windows or mac
            from PIL import Image
            output = output_folder.parent / f"{filepath.stem}.ico"
            largest = list(output_folder.glob("*_1024.png"))[0]
            img = Image.open(str(largest))
            icon_sizes = [(x, x) for x in self.valid_sizes]
            img.save(str(output), sizes=icon_sizes)


import re
from pprint import pprint
from more_itertools import split_at, split_before
from coldtype.geometry import Point, Line, Rect
from coldtype.sh.context import SHLookup, SHContext


SH_UNARY_SUFFIX_FUNCS = {
    "~": "reverse",
    "¶": "to_pen",
}

SH_UNARY_TO_STRING = {
    "⊢": "w",
    "⊣": "e",
    "⊤": "n",
    "⊥": "s",
    "⌶": "cx",
    "Ｈ": "cy",
    "←": "W",
    "↑": "N",
    "→": "E",
    "↓": "S",
    "↖": "NW",
    "↗": "NE",
    "↘": "SE",
    "↙": "SW",
    "•": "C",
}

SH_UNARY_SUFFIX_PROPS = {
    "⊢": "ew",
    "⊣": "ee",
    "⊤": "en",
    "⊥": "es",
    "⌶": "ecx",
    "Ｈ": "ecy",
    "←": "pw",
    "↑": "pn",
    "→": "pe",
    "↓": "ps",
    "↖": "pnw",
    "↗": "pne",
    "↘": "pse",
    "↙": "psw",
    "•": "pc",
    "⍺": "start",
    "⍵": "end",
    "µ": "mid",
}

SH_BINARY_OPS = {
    "I": "inset",
    "O": "offset",
    "C": "columns",
    "R": "rows",
    "@": "__getitem__",
    "↕": "extr",
    #"P": "project",
    "∏": "project",
    "π": "pinch",
    "†": "take_curve",
}

SH_BINARY_OPS_EDGEAWARE = {
    "T": "take",
    "S": "subtract",
    "E": "expand",
    "M": "maxima",
}

SH_JOINS = {
    "⨝": ["join"],
    "∩": ["intersection"],
    "∮": lambda a, b: f"DraftingPen().moveTo({a}.start).lineTo({a}.end).lineTo({b}.end).lineTo({b}.start).closePath()",
    "⫎": lambda a, b: f"Rect.FromPoints({a}, {b})"
}

SH_BACKREFS = {
    "〱": "_last",
}

SH_EXPLODES = {
    "〻": "_last",
}

SH_PATH_OPS = {
    "ɜ": "endPath",
    "ɞ": "closePath",
    "Я": "reverse"
}

def shchain(s):
    chars = list(SH_BINARY_OPS_EDGEAWARE.keys())
    chars.extend(SH_BINARY_OPS.keys())
    chars.extend(SH_UNARY_SUFFIX_PROPS)
    chars.extend(SH_UNARY_SUFFIX_FUNCS)
    chars.append(">")
    chars.append("Ƨ")
    
    cs = ["".join(x) for x in split_before(s, lambda x: x in chars) if x[0] != ">"]
    out = cs[0]
    spre = re.compile(",|—")
    skip = False

    for c in cs[1:]:
        f = c[0]
        if f == "Ƨ":
            skip = True
            continue
        elif skip:
            skip = False
            continue

        if f in SH_BINARY_OPS:
            fn = SH_BINARY_OPS[f]
            d = None
            if c[1] in ["X", "Y"]:
                d = c[1]
                args = spre.split(c[2:])
            else:
                args = spre.split(c[1:])
            if d:
                fn += "_" + d.lower()
            for i, a in enumerate(args):
                if a == "auto" or a == "a":
                    args[i] = '"auto"'
            out += f".{fn}({','.join(args)})"
        elif f in SH_BINARY_OPS_EDGEAWARE:
            fn = SH_BINARY_OPS_EDGEAWARE[f]
            d = "XY"
            if c[1] in ["X", "Y"]:
                d = c[1]
                args = spre.split(c[2:])
            else:
                args = spre.split(c[1:])
            for i, a in enumerate(args):
                if a[0] == "-":
                    e = "mn"
                elif a[0] == "=":
                    e = "md"
                elif a[0] == "+":
                    e = "mx"
                else:
                    raise Exception("Edge not matched", args[0])
                if d == "XY":
                    args[i] = (a[1:], '"'+e+"xy"[i]+'"')
                else:
                    args[i] = (a[1:], '"'+e+d.lower()+'"')
                out += f".{fn}({','.join(args[i])})"
        elif f in SH_UNARY_SUFFIX_PROPS:
            fn = SH_UNARY_SUFFIX_PROPS[f]
            out += f".{fn}" #+ c[1:]
        elif f in SH_UNARY_SUFFIX_FUNCS:
            fn = SH_UNARY_SUFFIX_FUNCS[f]
            out += f".{fn}()" #+ c[1:]
    
    return out

def shterm(s:str):
    return shchain(s)

def shphrase(s):
    terms = []
    splits = list(SH_JOINS.keys())
    for idx, _t in enumerate(split_at(s, lambda x: x in splits, keep_separator=1)):
        t = "".join(_t)
        if idx % 2 == 0:
            terms.append("("+shterm(t)+")")
        else:
            terms.append(t)

    out = ""
    t1 = terms[0]
    i = 1
    if i == len(terms):
        return t1    
    else:
        while i < len(terms):
            op_s = terms[i]
            if op_s in SH_JOINS:
                op = SH_JOINS[op_s]
                t2 = terms[i+1]
                
                for k, v in SH_BACKREFS.items():
                    t2 = t2.replace(k, f"({t1})")
                
                if callable(op):
                    out += op(t1, t2)
                else:
                    out += f"({t1}.{op[0]}({t2}))"
                i += 2
    return out

def shgroup(s):
    if s.startswith("Ƨ"):
        return None

    s = s.replace("(", "[").replace(")", "]")
    rg = re.compile(r"\[([^\]]+)\]")

    def expand(m):
        return f"({shphrase(m.group(1))})"

    rm = rg.findall(s)
    while len(rm) > 0:
        s = rg.sub(expand, s)
        rm = rg.findall(s)
    
    return shphrase(s)

def sh(s, ctx:SHContext=None, dps=None, subs={}):
    from coldtype.pens.draftingpen import DraftingPen

    #print("SH>", s, subs)

    if ctx is None:
        ctx = SHContext()

    evaled = []
    last_locals = {**ctx.locals}
    s = s.replace("_", "")
    s = "ƒ"+re.sub(r"[\s\n]+", "ƒ", s).strip()

    def expand_multisuffix(m):
        out = []
        arrows = list(m.group(2))
        for a in arrows:
            out.append(m.group(1)+a)
        return "ƒ".join(out)
    
    def do_eval(phrase):
        py = (shgroup(phrase))
        if not py:
            return None

        for k, v in SH_PATH_OPS.items():
            py = py.replace(k, '"' + v + '"')
        
        for k, v in ctx.lookups.items():
            py = py.replace(v.symbol, f"ctx.{k}.")

        for k, v in ctx.subs.items():
            py = py.replace(k, v(ctx) if callable(v) else v)
        
        for k, v in subs.items():
            py = py.replace(k, str(v))

        #print("EVAL<", py)

        try:
            res = eval(py, dict(
                ctx=ctx,
                _last=evaled[-1] if len(evaled) > 0 else None,
                _dps=dps,
                Point=Point,
                Line=Line,
                Rect=Rect,
                DraftingPen=DraftingPen)
                , last_locals)
            #print("LOCALS", last_locals)
            return res
        except SyntaxError as e:
            print("SYNTAX ERROR", e, phrase, py)
            return None

    #s = re.sub(r"([\$\&]{1}[a-z]+)([↖↑↗→↘↓↙←•⍺⍵µ]{2,})", expand_multisuffix, s)

    # for k, v in SH_PATH_OPS.items():
    #     s = s.replace(k, '"' + v + '"')

    join_to_path = False
    splits = ["ƒ"]
    splits.extend(SH_EXPLODES.keys())

    s = re.sub("ƒ\-[^ƒ]+", "", s)

    for phrase in split_before(s, lambda x: x in splits):
        phrase = "".join(phrase).strip()
        #print("PHRASE", phrase)
        last = None
        if not phrase:
            continue
        if phrase[0] in SH_EXPLODES:
            phrase = "_last"+phrase[1:]
        #    last = evaled[-1]
        if phrase[0] == "ƒ":
            phrase = phrase[1:]
        if not phrase:
            continue

        if phrase == "∫":
            phrase = "'∫'"

        more = []
        if "ø" in phrase:
            phrase = phrase.replace("ø", "")
        elif "|" in phrase:
            tuple = phrase.split("|")
            for i, t in enumerate(tuple):
                if isinstance(t, str):
                    if "∑" in t:
                        t = ",".join([f"'{c}'" for c in t])
                    elif len(t) > 1:
                        if t[0] in SH_UNARY_TO_STRING:
                            tuple[i] = [SH_UNARY_TO_STRING[x] for x in t]
                            continue
                    else:
                        if t in SH_UNARY_TO_STRING:
                            tuple[i] = SH_UNARY_TO_STRING[t]
                            continue
                tuple[i] = do_eval(t)
            more = tuple
            phrase = tuple[-1]

        if more:
            evaled.append(more)
        else:
            evaled.append(do_eval(phrase))
        if dps is not None:
            dps.append(evaled[-1])
    
    ctx.locals = {**ctx.locals, **last_locals}
    return evaled


from typing import Any, NamedTuple, Optional, Sequence, Tuple, Union
from fontTools.misc.arrayTools import unionRect


Number = Union[int, float]
Rectangle = Tuple[Number, Number, Number, Number]  # xMin, yMin, xMax, yMax


class RectTree(NamedTuple):

    """Given a sorted list of (rectangle, object) items, build a tree structure
    that allows to efficiently find objects that overlap with a target rectangle.

    Use the RectTree.fromSeq(seq) class method to build a tree.

    The tree.iterIntersections(targetRect) method iterates over the set of objects
    that overlap with targetRect (in order of the original sequence).

    The tree.firstIntersection(targetRect) method returns the first overlapping
    object or None.

    This implementation is targeted towards a more or less one-dimensional layout
    of the objects, for example a line of glyphs. The direction of the layout is
    not important, but it's most efficient to sort the objects along the intended
    direction. For example, a sequence of glyph bounding boxes that are layed out
    horizontally should be sorted horizontally (although right to left or left to
    right won't make a difference). The output of, say, hb-shape will do just
    fine, regardless of whether the layout is horizontal or vertical.

    Rectangles here have the form (xMin, yMin, xMax, yMax).
    """

    bounds: Rectangle
    leaf: Any
    left: Optional["RectTree"]
    right: Optional["RectTree"]

    @classmethod
    def fromSeq(cls, seq: Sequence[Tuple[Rectangle, Any]]):
        if len(seq) == 0:
            # empty tree, pass None for bounds as a special case
            return cls(None, None, None, None)
        elif len(seq) == 1:
            bounds, leaf = seq[0]
            return cls(bounds, leaf, None, None)
        mid = len(seq) // 2
        left = cls.fromSeq(seq[:mid])
        right = cls.fromSeq(seq[mid:])
        bounds = unionRect(left[0], right[0])
        return cls(bounds, None, left, right)

    def iterIntersections(self, targetBounds: Rectangle):
        if self.bounds is None:
            # empty tree
            return
        if not hasIntersection(self.bounds, targetBounds):
            return
        if self.left is None:
            assert self.right is None
            yield self.leaf
        else:
            yield from self.left.iterIntersections(targetBounds)
            yield from self.right.iterIntersections(targetBounds)

    def firstIntersection(self, targetBounds: Rectangle, default=None):
        return next(self.iterIntersections(targetBounds), default)


def hasIntersection(rect1, rect2):
    """Return a boolean. If the input rectangles intersect, return
    True, return False if the input rectangles do not intersect.
    """
    (xMin1, yMin1, xMax1, yMax1) = rect1
    (xMin2, yMin2, xMax2, yMax2) = rect2
    return ((xMin1 < xMax2 and xMax1 > xMin2) and (yMin1 < yMax2 and yMax1 > yMin2))



import json, re
from pathlib import Path
from coldtype.time.sequence import Sequence, ClipTrack, Clip
from coldtype.time.audio import Wavfile
from coldtype.text import Style, StyledString
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.pens.dattext import DATText
from coldtype.geometry import Rect
from coldtype.color import hsl, bw
from coldtype.renderable.animation import animation
from coldtype.renderable import Overlay
from contextlib import contextmanager


class Subtitler(Sequence):
    def __init__(self, path, audio_path, duration=None, fps=30, storyboard=[0]):
        self.path = Path(path).expanduser().absolute()
        self.path.parent.mkdir(exist_ok=True, parents=True)
        
        self.audio_path = Path(audio_path).expanduser().absolute()
        self.wavfile = Wavfile(self.audio_path)

        if duration is None:
            duration = self.wavfile.framelength
        
        try:
            self.data = json.loads(self.path.read_text())
        except (FileNotFoundError, json.JSONDecodeError):
            self.data = dict(workarea_track=0, tracks=[dict(clips=[])])
        
        tracks = []
        for tidx, t in enumerate(self.data.get("tracks")):
            clips = []
            for cidx, c in enumerate(t.get("clips")):
                cs = []
                txt = c.get("text")
                start = c.get("start")
                end = c.get("end")
                if "≈" in txt[1:]:
                    for lidx, ltxt in enumerate(txt.split("≈")):
                        if lidx > 0:
                            ltxt = "≈" + ltxt
                        cs.append(Clip(ltxt, start, end, cidx, tidx))
                else:
                    cs.append(Clip(txt, start, end, cidx, tidx))
                clips.extend(cs)
            tracks.append(ClipTrack(self, clips, []))
        
        super().__init__(duration, fps, storyboard, tracks, workarea_track=self.data["workarea_track"])

        if len(self.data["tracks"]) <= self.workarea_track:
            self.data["tracks"].append(dict(clips=[]))
            self.tracks.append(ClipTrack(self, [], []))
    
    def persist(self):
        self.path.write_text(json.dumps(self.data, indent=2, ensure_ascii=False), encoding="utf-8")
    
    def conform(self, clips):
        clips = sorted(clips, key=lambda c: c["start"])
        for cidx, clip in enumerate(clips):
            try:
                next_clip = clips[cidx+1]
            except IndexError:
                next_clip = None
            if next_clip:
                if clip["end"] > next_clip["start"]:
                    clip["end"] = next_clip["start"]
        return clips
    
    def clips(self):
        clips = self.data["tracks"][self.workarea_track]["clips"]
        for cidx, c in enumerate(clips):
            c["idx"] = cidx
        return clips
    
    def text_for_frame(self, fi):
        clips = [c for c in self.clips() if c.get("start") == fi]
        if len(clips) > 0:
            return clips[0].get("text")
        else:
            return ""
    
    def remove_clip_at_frame(self, fi, clips):
        return [c for c in clips if c["start"] != fi]
    
    def overwrite_clips(self, clips):
        self.data["tracks"][self.workarea_track]["clips"] = self.conform(clips)
        self.persist()
    
    def delete_clip(self, fi, clips):
        clips = self.remove_clip_at_frame(fi, clips)
        self.overwrite_clips(clips)
    
    def add_clip(self, fi, text, clips):
        clips = self.remove_clip_at_frame(fi, clips)
        _, matches = self.closest(fi, +1, clips)
        end = None
        for (pos, cidx) in matches:
            if pos == "start":
                end = clips[cidx]["start"]
        if end is None:
            end = self.duration
        clips.append(dict(text=text, start=fi, end=end))
        self.overwrite_clips(clips)
    
    @contextmanager
    def active_clip(self, fi, clips):
        cidx = self.current_clip_idx(fi, clips)
        if cidx > -1:
            try:
                yield clips[cidx]
            finally:
                self.overwrite_clips(clips)
    
    def cut_clip(self, fi, clips):
        with self.active_clip(fi, clips) as c:
            c["end"] = fi
        
    def mod_clip_text(self, fi, clips, fn):
        with self.active_clip(fi, clips) as c:
            ctxt = c["text"]
            csym = None
            if ctxt[0] in ["≈", "*", "+"]:
                csym = ctxt[0]
                ctxt = ctxt[1:]
            c["text"] = fn(csym, ctxt)

    def closest(self, fi, direction, clips, ignore_on=False):
        closest = []
        closest_cut = 5000
        closest_frame = None

        for c in clips:
            if direction < 0:
                sc = fi - c["start"]
                ec = fi - c["end"]
            elif direction > 0:
                sc = c["start"] - fi
                ec = c["end"] - fi
            
            if sc >= 0 and sc < closest_cut:
                if sc == 0 and ignore_on:
                    pass
                else:
                    closest_cut = sc
                    closest_frame = c["start"]
            if ec >= 0 and ec < closest_cut:
                if ec == 0 and ignore_on:
                    pass
                else:
                    closest_cut = ec
                    closest_frame = c["end"]
        
        for c in clips:
            if c["start"] == closest_frame:
                closest.append(["start", c["idx"]])
            elif c["end"] == closest_frame:
                closest.append(["end", c["idx"]])
        
        if not ignore_on:
            if closest_frame == fi and len(closest) == 1:
                return self.closest(fi, direction, clips, ignore_on=True)
        
        return closest_frame == fi, closest
    
    def closest_to_playhead(self, fi, which, clips):
        on_cut, matches = self.closest(fi, which, clips)
        if not on_cut:
            for (pos, cidx) in matches:
                clips[cidx][pos] = fi
        self.overwrite_clips(clips)
    
    # def prev_and_next(self, fi):
    #     clips = self.clips(self.workarea_track)
    #     curr_clip = None
    #     next_clip = None
    #     prev_clip = None
        
    #     for cidx, clip in enumerate(clips):
    #         if curr_clip and not next_clip:
    #             if clip["start"] > fi:
    #                 next_clip = clip
            
    #         if clip["end"] <= fi:
    #             prev_clip = clip
            
    #         if clip["start"] <= fi < clip["end"]:
    #             curr_clip = clip
        
    #     return prev_clip, curr_clip, next_clip, clips

    def current_clip_idx(self, fi, clips):
        for clip in clips:
            if clip["start"] <= fi < clip["end"]:
                return clip["idx"]
        return -1
    
    # def prev(self, clip, clips):
    #     for cidx, c in enumerate(clips):
    #         if c == clip:
    #             try:
    #                 return clips[cidx-1]
    #             except IndexError:
    #                 return None
    
    # def next(self, clip, clips):
    #     for cidx, c in enumerate(clips):
    #         if c == clip:
    #             try:
    #                 return clips[cidx+1]
    #             except IndexError:
    #                 return None

    def process_state(self, rs, fi, clips=None):
        if not clips:
            clips = self.clips()

        if rs.text:
            self.add_clip(fi, rs.text, clips)
        elif rs.cmd:
            if rs.cmd == "d":
                self.delete_clip(fi, clips)
            elif rs.cmd == "c":
                self.cut_clip(fi, clips)
            elif rs.cmd == "p":
                self.persist()
            elif rs.cmd == "q":
                #print(self.closest(fi, -1, clips))
                self.closest_to_playhead(fi, -1, clips)
            elif rs.cmd == "w":
                #print(self.closest(fi, +1, clips))
                self.closest_to_playhead(fi, +1, clips)
            elif rs.cmd == "ts":
                self.mod_clip_text(fi, clips,
                    lambda s, t: "*" + t if s != "*" else t)
            elif rs.cmd == "tl":
                self.mod_clip_text(fi, clips,
                    lambda s, t: "≈" + t if s != "≈" else t)
            elif rs.cmd == "tc":
                self.mod_clip_text(fi, clips,
                    lambda s, t: "+" + t if s != "+" else t)
            elif re.match(r"[0-9]{1}", rs.cmd):
                self.data["workarea_track"] = int(rs.cmd)
                self.persist()
    
    def clip_timeline(self, fi, far, sw=30, sh=50, font_name=None):
        seq = DATPens()

        seq += (DATPen()
            .rect(Rect(0, 0, (self.duration * sw), sh)).f(bw(0.5, 0.35)))

        for tidx, t in enumerate(self.tracks):
            for cgidx, cg in enumerate(t.clip_groups):
                for cidx, c in enumerate(cg.clips):
                    r = Rect(c.start*sw, sh*tidx, (c.duration*sw)-2, sh)
                    current = c.start-25 <= fi <= c.end+25
                    if current:
                        seq.append(DATPens([
                            (DATPen()
                                .rect(r)
                                .f(hsl(cgidx*0.4+tidx*0.2, 0.75).lighter(0.2))),
                            (DATText(c.input_text,
                                Style(font_name, 24, load_font=0, fill=bw(0, 1 if tidx == self.workarea_track else 0.5)),
                                r.inset(3, 8)))]))

        on_cut = fi in self.jumps()

        seq.translate(far.w/2 - fi*sw, 10)

        seq.append(DATPen()
            .rect(far.take(sh*len(self.tracks)+20, "mny").take(6 if on_cut else 2, "mdx"))
            .f(hsl(0.9, 1) if on_cut else hsl(0.8, 0.5)))
        
        return seq


class lyric_editor(animation):
    def __init__(self, timeline:Subtitler, rect=(1080, 1080), data_font="Times", **kwargs):
        self.data_font = data_font
        super().__init__(
            rect=rect,
            timeline=timeline,
            audio=timeline.audio_path,
            rstate=1,
            watch=[timeline.path],
            **kwargs)

    def runpost(self, res, rp, rs):
        res = super().runpost(res, rp, rs)
        f:Frame = rp.args[0]
        if Overlay.Timeline in rs.overlays:
            self.timeline.process_state(rs, f.i)
            seq = self.timeline.clip_timeline(f.i, f.a.r, font_name=self.data_font)
            return DATPens([res, seq])
        return res

if __name__ == "<run_path>":
    from coldtype.test import recmono, co

    tl = Subtitler(
        #"test/lyric_data.json",
        #"test/media/helloworld.wav",
        "test/lyric_data2.json",
        "~/Type/fouchet/de/media/jitterbugging.wav",
        storyboard=[19])

    @lyric_editor(tl, bg=0, data_font=recmono)
    def lyric_entry2(f, rs):
        def render_clip_fn(f, idx, clip, ftext):
            if "stylized" in clip.styles:
                return ftext.upper(), Style(co, 150, wdth=0.5, tu=50, fill=hsl(0.61, s=1, l=0.7))
            return ftext, Style(recmono, 100, fill=1)

        return (tl.clip_group(0, f.i)
            .pens(f, render_clip_fn, f.a.r.inset(50))
            #.f(1)
            #.xa()
            #.align(f.a.r)
            .remove_futures())


import math
import easing_functions as ef
from fontTools.misc.bezierTools import splitCubic, splitLine

from typing import List

eases = dict(
    cei=ef.CubicEaseIn,
    ceo=ef.CubicEaseOut,
    ceio=ef.CubicEaseInOut,
    qei=ef.QuadEaseIn,
    qeo=ef.QuadEaseOut,
    qeio=ef.QuadEaseInOut,
    eei=ef.ExponentialEaseIn,
    eeo=ef.ExponentialEaseOut,
    eeio=ef.ExponentialEaseInOut,
    sei=ef.SineEaseIn,
    seo=ef.SineEaseOut,
    seio=ef.SineEaseInOut,
    bei=ef.BounceEaseIn,
    beo=ef.BounceEaseOut,
    beio=ef.BounceEaseInOut,
    eleo=ef.ElasticEaseOut,
    elei=ef.ElasticEaseIn,
    elieo=ef.ElasticEaseInOut,
    eleio=ef.ElasticEaseInOut)


def curve_pos_and_speed(curve, x):
    x1000 = x*1000
    for idx, (action, pts) in enumerate(curve.value):
        if action in ["moveTo", "endPath", "closePath"]:
            continue
        last_action, last_pts = curve.value[idx-1]
        if action == "curveTo":
            o = -1
            a = last_pts[-1]
            b, c, d = pts
            if x1000 == a[0]:
                o = a[1]/1000
                eb = a
                ec = b
            elif x1000 == d[0]:
                o = d[1]/1000
                eb = c
                ec = d
            elif x1000 > a[0] and x1000 < d[0]:
                e, f = splitCubic(a, b, c, d, x1000, isHorizontal=False)
                ez, ea, eb, ec = e
                o = ec[1]/1000
            else:
                continue
            tangent = math.degrees(math.atan2(ec[1] - eb[1], ec[0] - eb[0]) + math.pi*.5)
            #print(o, tangent)
            if tangent >= 90:
                t = (tangent - 90)/90
            else:
                t = tangent/90
            if o != -1:
                return o, t
    raise Exception("No curve value found!")


def ease(style, x):
    """
    Though available as a general-purpose function, this logic is usually accessed through something like the `.progress` function on an animation or timeable.

    Return two values — the first is the easing result at a given time x; the second is the tangent to that, if calculable (is not, atm, calculable for the mnemonics given)

    for reference, easing mnemonics:

    * cei = CubicEaseIn
    * ceo = CubicEaseOut
    * ceio = CubicEaseInOut
    * qei = QuadEaseIn
    * qeo = QuadEaseOut
    * qeio = QuadEaseInOut
    * eei = ExponentialEaseIn
    * eeo = ExponentialEaseOut
    * eeio = ExponentialEaseInOut
    * sei = SineEaseIn
    * seo = SineEaseOut
    * seio = SineEaseInOut
    * bei = BounceEaseIn
    * beo = BounceEaseOut
    * beio = BounceEaseInOut
    * eleo = ElasticEaseOut
    * elei = ElasticEaseIn,
    * eleio = ElasticEaseInOut
    """
    if style == "linear" or style == "lin" or style == "l":
        return x, 0.5
    e = eases.get(style)
    if e:
        return e().ease(x), 0.5
    elif hasattr(style, "moveTo"):
        return curve_pos_and_speed(style, x)
    elif False:
        if style in easer_ufo:
            return curve_pos_and_speed(DATPen().glyph(easer_ufo[style]), x)
        else:
            raise Exception("No easing function with that mnemonic")
    else:
        raise Exception("No easing function with that mnemonic")